using Microsoft.Data.Sqlite;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace Elwig.Helpers.Billing {
    public class Billing {

        protected readonly int Year;
        protected readonly AppDbContext Context;
        protected readonly Dictionary<string, string> Attributes;
        protected readonly Dictionary<string, (decimal?, decimal?)> Modifiers;
        protected readonly Dictionary<string, (string, string?, string?, string?, int?, int?, decimal?)> AreaComTypes;

        public Billing(int year) {
            Year = year;
            Context = new AppDbContext();
            Attributes = Context.WineAttributes.ToDictionary(a => a.AttrId, a => a.Name);
            Modifiers = Context.Modifiers.Where(m => m.Year == Year).ToDictionary(m => m.ModId, m => (m.Abs, m.Rel));
            AreaComTypes = Context.AreaCommitmentTypes.ToDictionary(v => v.VtrgId, v => (v.SortId, v.AttrId1, v.AttrId2, v.Discriminator, v.MinKgPerHa, v.MaxKgPerHa, v.PenaltyAmount));
        }

        public async Task FinishSeason() {
            using var cnx = await AppDbContext.ConnectAsync();
            using (var cmd = cnx.CreateCommand()) {
                cmd.CommandText = $"""
                    UPDATE season
                    SET (start_date, end_date) = (SELECT MIN(date), MAX(date) FROM delivery WHERE year = {Year}),
                        bin_1_name = 'gebunden mit zwei Attributen',
                        bin_2_name = 'gebunden mit (erstem) Attribut',
                        bin_3_name = 'gebunden mit zweitem Attribut',
                        bin_4_name = 'gebunden ohne Attribut',
                        bin_5_name = 'ungebunden',
                        bin_6_name = NULL,
                        bin_7_name = NULL,
                        bin_8_name = NULL,
                        bin_9_name = NULL
                    WHERE year = {Year}
                    """;
                await cmd.ExecuteNonQueryAsync();
            }
        }

        public async Task CalculateBins() {
            using var cnx = await AppDbContext.ConnectAsync();
            var memberOblRig = await GetMemberRightsObligations(cnx, Year);
            var inserts = new List<(int, int, int, int, int, int, int)>();

            var deliveries = new List<(int, int, int, string, int, double, string, string[], string[])>();
            using (var cmd = cnx.CreateCommand()) {
                cmd.CommandText = $"""
                    SELECT mgnr, did, dpnr, sortid, weight, kmw, qualid, attributes, modifiers
                    FROM v_delivery
                    WHERE year = {Year}
                    ORDER BY mgnr, sortid, abgewertet ASC, LENGTH(attributes) DESC, COALESCE(attributes, '~'), kmw DESC, lsnr, dpnr
                    """;
                var reader = await cmd.ExecuteReaderAsync();
                while (await reader.ReadAsync()) {
                    deliveries.Add((
                        reader.GetInt32(0), reader.GetInt32(1), reader.GetInt32(2), reader.GetString(3), reader.GetInt32(4),
                        reader.GetDouble(5), reader.GetString(6),
                        reader.IsDBNull(7) ? Array.Empty<string>() : reader.GetString(7).Split(",").Order().ToArray(),
                        reader.IsDBNull(8) ? Array.Empty<string>() : reader.GetString(8).Split(",").Order().ToArray()
                    ));
                }
            }

            int lastMgNr = 0;
            Dictionary<string, int>? rights = null;
            foreach (var (mgnr, did, dpnr, sortid, weight, kmw, qualid, attributes, modifiers) in deliveries) {
                if (lastMgNr != mgnr) {
                    rights = memberOblRig.GetValueOrDefault(mgnr, (new(), new())).Item2;
                }
                if (rights == null || rights.Count == 0 || qualid == "WEI" || qualid == "RSW" || qualid == "LDW") {
                    // Mitglied hat keine Flächenbindungen, oder
                    // Nicht mindestens Qualitätswein (QUW) -> ungebunden
                    inserts.Add((did, dpnr, 0, 0, 0, 0, weight));
                    continue;
                }

                if (attributes.Length > 2)
                    throw new NotSupportedException();

                int w = weight;
                int[] b = new int[4];
                foreach (var p in Utils.Permutate(attributes)) {
                    var c = p.Count();
                    var key = sortid + string.Join("", p);
                    if (rights.ContainsKey(key)) {
                        int i = 0;
                        if (c == 1) {
                            i = (p.ElementAt(0) == attributes[0]) ? 1 : 2;
                        } else if (c == 0) {
                            i = b.Length - 1;
                        }
                        var v = Math.Max(0, Math.Min(rights[key], w));
                        b[i] += v;
                        rights[key] -= v;
                        w -= v;
                    }
                }
                inserts.Add((did, dpnr, b[0], b[1], b[2], b[3], weight - b[0] - b[1] - b[2] - b[3]));
                lastMgNr = mgnr;
            }

            using (var cmd = cnx.CreateCommand()) {
                cmd.CommandText = $"""
                    INSERT INTO delivery_part_bin (year, did, dpnr, bin_1, bin_2, bin_3, bin_4, bin_5)
                    VALUES {string.Join(",\n       ", inserts.Select(i => $"({Year}, {i.Item1}, {i.Item2}, {i.Item3}, {i.Item4}, {i.Item5}, {i.Item6}, {i.Item7})"))}
                    ON CONFLICT DO UPDATE
                    SET bin_1 = excluded.bin_1,
                        bin_2 = excluded.bin_2,
                        bin_3 = excluded.bin_3,
                        bin_4 = excluded.bin_4,
                        bin_5 = excluded.bin_5,
                        bin_6 = NULL,
                        bin_7 = NULL,
                        bin_8 = NULL,
                        bin_9 = NULL;
                    """;
                await cmd.ExecuteNonQueryAsync();
            }
        }

        public static async Task<Dictionary<int, (Dictionary<string, int>, Dictionary<string, int>)>> GetMemberRightsObligations(SqliteConnection cnx, int year, int? mgnr = null) {
            var members = new Dictionary<int, (Dictionary<string, int>, Dictionary<string, int>)>();

            using var cmd = cnx.CreateCommand();
            cmd.CommandText = $"""
                SELECT mgnr, t.vtrgid,
                       SUM(COALESCE(area * min_kg_per_ha, 0)) / 10000 AS min_kg,
                       SUM(COALESCE(area * max_kg_per_ha, 0)) / 10000 AS max_kg
                FROM area_commitment c
                JOIN area_commitment_type t ON t.vtrgid = c.vtrgid
                WHERE ({(mgnr == null ? "NULL" : mgnr)} IS NULL OR mgnr = {(mgnr == null ? "NULL" : mgnr)}) AND
                      (year_from IS NULL OR year_from <= {year}) AND
                      (year_to IS NULL OR year_to >= {year})
                GROUP BY mgnr, t.vtrgid
                ORDER BY LENGTH(t.vtrgid) DESC, t.vtrgid
                """;

            var reader = await cmd.ExecuteReaderAsync();
            while (await reader.ReadAsync()) {
                var m = reader.GetInt32(0);
                var vtrgid = reader.GetString(1);
                if (!members.ContainsKey(m)) members[m] = (new(), new());
                members[m].Item1[vtrgid] = reader.GetInt32(2);
                members[m].Item2[vtrgid] = reader.GetInt32(3);
            }

            return members;
        }

        public static async Task<(Dictionary<string, int>, Dictionary<string, int>)> GetMemberRightsObligations(SqliteConnection cnx, int year, int mgnr) {
            var members = await GetMemberRightsObligations(cnx, year, (int?)mgnr);
            return members[mgnr];
        }

        public static async Task<Dictionary<string, int>> GetMemberBinWeights(int mgnr, int year, SqliteConnection cnx) {
            var bins = new Dictionary<string, int>();

            using var cmd = cnx.CreateCommand();
            cmd.CommandText = $"""
                SELECT bin, weight
                FROM v_delivery_bin
                WHERE (year, mgnr) = ({year}, {mgnr})
                """;

            var reader = await cmd.ExecuteReaderAsync();
            while (await reader.ReadAsync()) {
                var bin = reader.GetString(0);
                bins[bin] = reader.GetInt32(1);
            }

            return bins;
        }
    }
}
